@@ -1,8 +1,11 @@
-from dataclasses import dataclass
-from unittest.mock import Mock
+from typing import List
 
 import pytest
 
+import threading
+from dataclasses import dataclass
+from unittest.mock import Mock
+
 from autowired import (
     autowired,
     cached_property,
@@ -17,8 +20,8 @@ from autowired import (
     NotProvidedException,
     IllegalContextClass,
     MissingTypeAnnotation,
-    ProviderConflictException,
     provided,
+    thread_local_cached_property,
 )
 
 
@@ -117,6 +120,28 @@ def test_unresolvable_dependency():
         print(ctx.service4)
 
 
+def test_context_provider_names():
+    class TestContext(Context):
+        service1: Service1 = autowired()
+        my_service_2: Service2 = autowired()
+
+    ctx = TestContext()
+    providers = ctx.container.get_providers()
+    assert len(providers) == 2
+
+    assert "service1" in [c.get_name() for c in providers]
+    assert "my_service_2" in [c.get_name() for c in providers]
+
+    class PascalCaseToSnakeCase:
+        pass
+
+    container = Container()
+    container.add(PascalCaseToSnakeCase())
+    providers = container.get_providers()
+    assert len(providers) == 1
+    assert providers[0].get_name() == "pascal_case_to_snake_case"
+
+
 def test_ambiguous_dependency():
     class TestContext(Context):
         service1_a: Service1 = autowired()
@@ -217,9 +242,30 @@ def test_unregister_dependency():
 
 def test_conflicting_bean():
     container = Container()
-    container.add(Provider.from_instance(Service0()))
-    with pytest.raises(ProviderConflictException):
-        container.add(Provider.from_instance(Service0()))
+
+    class Component:
+        pass
+
+    container.add(Component())
+    container.add(Component())
+
+    named = Component()
+    container.add(Provider.from_instance(named, "named_component"))
+
+    with pytest.raises(AmbiguousDependencyException):
+        container.resolve(Component)
+
+    with pytest.raises(AmbiguousDependencyException):
+        container.resolve(Dependency("component", Component))
+
+    # should work with exact name
+    container.resolve(Dependency("named_component", Component))
+
+    container.add(Provider.from_instance(Component(), "named_component"))
+
+    # should not work with multiple components with the same name
+    with pytest.raises(AmbiguousDependencyException):
+        container.resolve(Dependency("named_component", Component))
 
 
 def test_use_correct_subtype():
@@ -517,16 +563,51 @@ def test_transient_autowired_field():
     assert id(ctx.service_a) == id(ctx.service_a)
 
 
+def test_untyped_component_arg():
+    class ServiceA:
+        def __init__(self, arg):
+            self.arg = arg
+
+    container = Container()
+
+    with pytest.raises(UnresolvableDependencyException):
+        container.autowire(ServiceA)
+
+
+def test_untyped_component_arg_with_default():
+    class SomeClass:
+        pass
+
+    default_value = SomeClass()
+
+    class SomeService:
+        def __init__(self, arg=default_value):
+            self.arg = arg
+
+    container = Container()
+
+    not_default_value = SomeClass()
+
+    container.add(not_default_value)
+    service = container.resolve(SomeService)
+
+    assert isinstance(service.arg, SomeClass)
+    assert service.arg is not default_value
+    assert service.arg is not_default_value
+
+
 def test_inherited_context():
     class ServiceA:
-        pass
+        def __init__(self, arg: str = ""):
+            self.arg = arg
+
+    class ServiceB:
+        def __init__(self, arg: str = ""):
+            self.arg = arg
 
     class BaseContext(Context):
         service_a: ServiceA = autowired()
 
-    class ServiceB:
-        pass
-
     class ExtendedContext(BaseContext):
         service_b: ServiceB = autowired()
 
@@ -535,6 +616,18 @@ def test_inherited_context():
     assert isinstance(ctx.service_a, ServiceA)
     assert isinstance(ctx.service_b, ServiceB)
 
+    class ExtendedContext2(BaseContext):
+        service_a: ServiceA = autowired(arg="foo")
+
+        @cached_property
+        def service_b(self) -> ServiceB:
+            return self.autowire(ServiceB, arg="bar")
+
+    ctx2 = ExtendedContext2()
+
+    assert ctx2.service_a.arg == "foo"
+    assert ctx2.service_b.arg == "bar"
+
 
 def test_provider_from_supplier():
     instance = Service0()
@@ -601,3 +694,188 @@ def test_provider_factory_methods():
         Provider.from_supplier(supplier_typed).get_instance(Mock(), Mock()),
         Service0,
     )
+
+
+def test_thread_local_property():
+    import threading
+
+    class ServiceA:
+        def __init__(self):
+            self.thread_id = threading.get_ident()
+
+    class ServiceB:
+        def __init__(self, service_a: ServiceA):
+            self.service_a = service_a
+            self.thread_id = threading.get_ident()
+
+    class TestContext(Context):
+        service_a: ServiceA = autowired()
+
+        @thread_local_cached_property
+        def service_b(self) -> ServiceB:
+            return self.autowire(ServiceB)
+
+    ctx = TestContext()
+    main_thread_service_b = ctx.service_b
+
+    assert isinstance(main_thread_service_b, ServiceB)
+    assert main_thread_service_b is ctx.service_b
+
+    def in_new_thread():
+        new_thread_service_b = ctx.service_b
+        assert new_thread_service_b is not main_thread_service_b
+        assert new_thread_service_b is ctx.service_b
+
+        assert new_thread_service_b.service_a is main_thread_service_b.service_a
+
+    _test_in_thread(in_new_thread)
+
+
+def _test_in_thread(func):
+    thread_error = None
+
+    def in_new_thread():
+        try:
+            func()
+        except Exception as e:
+            nonlocal thread_error
+            thread_error = e
+
+    thread = threading.Thread(target=in_new_thread)
+    thread.start()
+    thread.join()
+
+    if thread_error is not None:
+        raise thread_error
+
+
+def test_thread_local_autowired():
+    class TestContext(Context):
+        service1: Service1 = autowired(thread_local=True)
+
+    ctx = TestContext()
+    main_thread_service1 = ctx.service1
+    assert isinstance(main_thread_service1, Service1)
+    assert ctx.service1 is main_thread_service1
+
+    def in_new_thread():
+        assert isinstance(ctx.service1, Service1)
+        assert ctx.service1 is not main_thread_service1
+        assert ctx.service1 is ctx.service1
+
+    _test_in_thread(in_new_thread)
+
+
+def test_thread_safe_autowired_field():
+    class SlowInitService:
+        def __init__(self):
+            import time
+
+            time.sleep(0.1)
+
+    class MyContext(Context):
+        service: SlowInitService = autowired()
+
+    ctx = MyContext()
+
+    from_thread = None
+
+    def in_thread():
+        nonlocal from_thread
+        from_thread = ctx.service
+
+    thread = threading.Thread(target=in_thread)
+    thread.start()
+
+    from_main = ctx.service
+
+    thread.join()
+    assert from_main is from_thread
+
+
+def test_context_value_selector():
+    @dataclass
+    class ServiceA:
+        foo: str = "bar"
+
+    @dataclass
+    class ServiceB:
+        foo: str = "baz"
+
+    @dataclass
+    class ServiceC:
+        foo: object
+
+    @dataclass
+    class ServiceD:
+        foo: object
+
+    @dataclass
+    class InnerConfig:
+        foo: str
+
+    class Config:
+        def __init__(self, inner: InnerConfig):
+            self.inner = inner
+
+    class MyContext(Context):
+        config: Config = provided()
+        a: ServiceA = autowired(foo=config.inner.foo)
+        # noinspection PyUnresolvedReferences
+        b: ServiceB = autowired(foo=config.wrong)
+        c: ServiceC = autowired(foo=object())
+        d: ServiceD = autowired(foo=c.foo)
+
+        def __init__(self, config: Config):
+            self.config = config
+
+    ctx = MyContext(Config(InnerConfig("bar")))
+
+    assert ctx.a.foo == "bar"
+
+    with pytest.raises(AttributeError):
+        print(ctx.b)
+
+    assert ctx.d.foo is ctx.c.foo
+
+
+def test_context_value_selector_direct():
+    @dataclass
+    class TestService:
+        service1: Service1
+        foo: str
+
+    class TestContext(Context):
+        foo: str = provided()
+        service1: Service1 = autowired()
+        test_service: TestService = autowired(service1=service1, foo=foo)
+
+        def __init__(self):
+            self.foo = "bar"
+
+    ctx = TestContext()
+    assert isinstance(ctx.test_service, TestService)
+    assert ctx.test_service.service1 is ctx.service1
+    assert ctx.test_service.foo == "bar"
+
+
+def test_generic_dependency():
+    @dataclass
+    class SomeData:
+        pass
+
+    @dataclass
+    class Service:
+        data: List[SomeData]
+
+    class TestContext(Context):
+        service: Service = autowired()
+
+        @cached_property
+        def data(self) -> List[SomeData]:
+            return [SomeData()]
+
+    ctx = TestContext()
+    assert isinstance(ctx.service, Service)
+    for data in ctx.service.data:
+        assert isinstance(data, SomeData)
